<?php

namespace Verifone\Hosted\Controller\Standard;

use Magento\Framework\App\{CsrfAwareActionInterface, Request\InvalidRequestException, RequestInterface};
use Magento\Sales\Model\Order;
use Verifone\Hosted\Controller\Checkout;
use Verifone\Hosted\Helper\Verifone;
use VerifoneEcomAPI\ApiWrapper\Http\SimpleCurl;

class Webhook extends Checkout implements CsrfAwareActionInterface {
    protected $_simpleCurl;
    protected $_orderRepo;
    protected $_transactionBuilder;

    public function __construct(
        \Magento\Framework\App\Action\Context                           $context,
        \Magento\Customer\Model\Session                                 $customerSession,
        \Magento\Checkout\Model\Session                                 $checkoutSession,
        \Magento\Quote\Api\CartRepositoryInterface                      $quoteRepository,
        \Magento\Sales\Model\OrderFactory                               $orderFactory,
        \Psr\Log\LoggerInterface                                        $logger,
        \Verifone\Hosted\Model\Verifone                                 $paymentMethod,
        \Verifone\Hosted\Helper\Verifone                                $checkoutHelper,
        \Magento\Quote\Api\CartManagementInterface                      $cartManagement,
        \Magento\Framework\Controller\Result\JsonFactory                $resultJsonFactory,
        \Magento\Sales\Api\OrderRepositoryInterface                     $orderRepo,
        \Magento\Sales\Model\Order\Payment\Transaction\BuilderInterface $transactionBuilder
    ) {
        parent::__construct($context, $customerSession, $checkoutSession, $quoteRepository, $orderFactory, $logger, $paymentMethod, $checkoutHelper, $cartManagement, $resultJsonFactory);

        $this->_simpleCurl = new SimpleCurl();
        $this->_orderRepo = $orderRepo;
        $this->_transactionBuilder = $transactionBuilder;
    }

    public function execute() {
        $input = file_get_contents('php://input');
        if (!$input)
            $this->api_response_error('Unable to process data');

        $data = json_decode($input, true);
        if (json_last_error() !== JSON_ERROR_NONE) {
            $this->_logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': Unable to decode response from API');
            $this->api_response_error('Unable to process data');
        }

        $paymentMethod = $this->getPaymentMethod();

        if ($this->stringStartsWith($data['eventType'], 'Checkout'))
            try {
                $checkout = $this->getCheckoutHelper()->getCheckout($paymentMethod, $data['recordId']);
                $data['recordId'] = $checkout['transaction_id'];
            } catch (\Exception $exception) {
                $this->_logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': There was a request for an order: ' . $exception->getMessage());
                $this->api_response_error('Unable to process data');
            }

        $response = $this->getCheckoutHelper()->getTransaction($paymentMethod, $data['recordId']);
        $transactionId = $response['id'];

        if (!empty($response['merchant_reference'])) {
            $quote = $this->_quoteRepository->get($response['merchant_reference']);
        } elseif (!empty($response['id'])) {
            $quote = $this->_quoteRepository->get($response['id']);
        } else {
            $this->_logger->error(__CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': API did not sent the ID of the quote');
            $this->api_response_error('Unable to process data');
        }

        $order = $this->getCheckoutHelper()->getOrderByIncrementId($quote->getReservedOrderId());

        switch ($response['status']) {
            case Verifone::VERIFONE_TRANSACTION_CAPTURE_AUTHORISED:
            case Verifone::VERIFONE_TRANSACTION_CAPTURE_SETTLED:
            case Verifone::VERIFONE_TRANSACTION_AUTHORISED:
            case Verifone::VERIFONE_TRANSACTION_SALE_AUTHORISED:
            case Verifone::VERIFONE_TRANSACTION_SALE_SETTLED:
            case Verifone::VERIFONE_TRANSACTION_AUTHORISATION_AUTHORISED:
            case Verifone::VERIFONE_TRANSACTION_CARD_VERIFICATION_AUTHORISED:
                if (!$order) {
                    $quote->setCheckoutMethod(\Magento\Checkout\Model\Type\Onepage::METHOD_GUEST);
                    $code = $paymentMethod->getCode();
                    $quote->setPaymentMethod($code);
                    $quote->getPayment()->importData(['method' => $code]);
                    $this->_quoteRepository->save($quote);

                    try {
                        $this->_cartManagement->placeOrder(
                            $quote->getId(),
                            $quote->getPayment()
                        );
                        $order = $this->getOrder();
                    } catch (\Exception $e) {
                        $this->_logger->error($e->getMessage());
                        $this->_logger->error($e->getTraceAsString());
                        if (isset($this->messageManager)) {
                            $this->messageManager->addExceptionMessage($e, __('We can\'t place the order.'));
                        }
                    }
                }
                $this->getCheckoutSession()
                    ->setLastSuccessQuoteId($quote->getId())
                    ->setLastQuoteId($quote->getId())
                    ->setLastRealOrderId($order->getIncrementId())
                    ->setLastOrderId($order->getId())
                    ->setLastOrderStatus($order->getStatus());

                /** Because with a refund, we can receive a transaction event but also a checkout event having transaction_id the original successful transaction, check that we don't already have a refunded amount before set status to complete.
                 */
                if (!$this->_checkoutHelper->checkVerifoneTransactionIdInOrderLogHistory($transactionId, $order) && !($order->getTotalRefunded() ?? 0)) {
                    $payment = $order->getPayment();
                    $paymentMethod = $payment->getMethodInstance();
                    $this->_checkoutHelper->setUpVerifoneSettings($paymentMethod);

                    $order->setExtOrderId($response['id'])
                        ->addStatusToHistory(Order::STATE_COMPLETE, __('Order payment is completed with transaction ID %1', $transactionId), true)
                        ->setStatus($this->_checkoutHelper->getVerifoneSettings()->getOrderStatus())
                        ->setState($this->_checkoutHelper->getVerifoneSettings()->getOrderState())
                        ->save();

                    $payment->setAdditionalInformation(
                        [\Magento\Sales\Model\Order\Payment\Transaction::RAW_DETAILS => (array)$response,
                            'verifone_order_status' => 'approved',
                            'verifone_order_number' => $transactionId,
                            'verifone_payment_product' => $response['payment_product']
                        ]
                    );
                }
                break;

            case Verifone::VERIFONE_TRANSACTION_PENDING:
                $order->setExtOrderId($response['id'])
                    ->setStatus(Order::STATE_PENDING_PAYMENT)
                    ->setState('new')
                    ->addStatusToHistory(Order::STATE_PENDING_PAYMENT, __('The payment is not done yet (transaction is in PRE/FINAL AUTH status), use the CAPTURE button to fully/partial capture the amount.
                    Attempting to refund the order in part or in full, before capturing, will void the authorization and cancel the payment.
                    Order status changed to Pending.'))
                    ->save();
                break;

            case Verifone::VERIFONE_TRANSACTION_REFUND_AUTHORISED:
            case Verifone::VERIFONE_TRANSACTION_REFUND_REFUNDED:
            case Verifone::VERIFONE_TRANSACTION_REFUNDED:
            case Verifone::VERIFONE_TRANSACTION_REFUND_SETTLED:
                if (!$order) break;

                if ($this->_checkoutHelper->checkVerifoneTransactionIdInOrderLogHistory($transactionId, $order)) break;

                $order->setStatus(Order::STATE_CLOSED)
                    ->setState(Order::STATE_CLOSED)
                    ->addStatusToHistory(Order::STATE_CLOSED, __('Order payment is refunded with transaction ID %1', $transactionId), true);
                if (!empty($response['amount'])) {
                    $refunded = $order->getTotalRefunded() ?? 0;
                    $verifoneResponseAmount = (float)$response['amount'];
                    $totalRefunded = $refunded + $verifoneResponseAmount;
                    $totalRefunded = min($totalRefunded, $order->getGrandTotal());

                    $order->setTotalRefunded($totalRefunded);
                }
                $order->save();
                break;

            default:
                if (Verifone::VERIFONE_WEBHOOK_EVENT_TYPE_REFUND_DECLINED == $data['eventType']) {
                    if (!$order) break;

                    $verifoneResponseAmount = (float)$response['amount'];
                    $refunded = $order->getTotalRefunded() ?? 0;
                    $totalRefunded = $refunded - $verifoneResponseAmount;
                    $totalRefunded=max($totalRefunded,0);

                    $order->setTotalRefunded($totalRefunded);
                } else {
                    $order->setState(Order::STATE_CLOSED)
                        ->setStatus(Order::STATE_CLOSED)
                        ->addStatusToHistory('FAILED', __('Order status changed to failed') . ' ' . $data['eventType'] . ' Response status:' . $response['status'])
                        ->save();
                }
        }

        exit;
    }

    public function api_response_error($message = null) {
        $this->api_response($message, 500);
    }

    public function api_response($message = null, $statusCode = 200) {
        http_response_code($statusCode);
        echo $message;
        exit;
    }

    public function createCsrfValidationException(RequestInterface $request): ?InvalidRequestException {
        return null;
    }

    public function validateForCsrf(RequestInterface $request): ?bool {
        return true;
    }

    public function api_response_ok($message = null) {
        $this->api_response($message, 200);
    }

    public function stringStartsWith( $string, $needle ) {
        $len = strlen( $needle );

        return ( substr( $string, 0, $len ) === $needle );
    }

}
